#require(RdbiPgSQL)


########################################################################################################################
## System preferences
########################################################################################################################
log.file = "pgUtils.log"
do.log = TRUE
log.level="DEBUG"

createDBTable <- function(conn,name=NULL,attributes=NULL,data.types=NULL,options=NULL,references=NULL,references.options=NULL,no.clobber=TRUE,auto.pk=TRUE,no.transaction=FALSE,silent=FALSE){
	if(is.null(name)){
		stop("ERROR, please submit the desired table name!\n\n")
	}
	if(is.null(attributes)){
		stop("ERROR, no attributes (columns) for the table are defined... do you really want to create a table without columns? \n If not submit the column names by the attributes parameter.\n\n")
	}
	if(is.null(data.types)){
		#stop("ERROR, please submit the data.types that your columns have. e.g for attributes=c(atext,somenumbers),data.types=c(character,integer). \n\n")
		cat(paste("Setting the data types for all attributes to  ","TEXT","\n\n"))
		data.types <- rep("TEXT",length(attributes))
	}
	else{
		if(length(data.types)!=length(attributes)){
			if(is.character(data.types[1])){
				data.types <- rep(data.types[1],length(attributes))
				cat(paste("Setting the data types for all attributes to  ",data.types[1],"\n\n"))
			}
			else{
				data.types <- rep(rpgsql.data.type(data.types[1]),length(attributes))
				cat(paste("Setting the data types for all attributes to  ",rpgsql.data.type(data.types[1]),"\n\n"))
			}
		}
	}
	if(!no.transaction){
		dbSendQuery(conn,"BEGIN TRANSACTION;")
	}
	if(!silent){
		if (no.clobber) {
	# Bug fix by Markus Neteler. $relname appended
		if (name %in% dbListTables(conn)$relname)
			stop(paste("Table name", name, "exists in database"))
		} else {
		# Bug fix by Markus Neteler. $relname appended
			if (name %in% dbListTables(conn)$relname){
				res <- dbSendQuery(conn, "DROP TABLE", double.quote(name))
				if(do.log){
					loggit(res,call=paste("createDBTable: INFO: table already existed... DROP TABLE", double.quote(name)))
				}
			}
		}
	}
 
	query <- paste("CREATE TABLE",double.quote(name),"(")
	col.def <- NULL
	# the options for the columns
	if(is.null(options)){
		options <- rep("",length(attributes))
	}
	if(auto.pk){
		col.def <- (paste(auto.pk.name(name)," BIGSERIAL PRIMARY KEY , ",sep=""))
	}
	# creating a list of the foreign keys (references to other tables).
	if(!is.null(references)){
		if(!is.null(references.options)){
			if(length(references.options)!=length(references)){
				ro <- references.options[1]
				references.options <- matrix(ncol=length(references),nrow=1)
				references.options[,] <- ro
			}
		}
		if(is.null(references.options)){
				references.options <- matrix(ncol=length(references),nrow=1)
				references.options[,] <- "NOT NULL"
		}
		for(i in 1:length(references)){
			col.def <- paste(col.def,auto.fk.name(references[i])," int8 REFERENCES ",references[i]," ( ",auto.pk.name(references[i])," ) ",references.options[i],", ",sep="")
		}
	}
	# defining the attributes / columns of the database table
	for(i in 1:length(attributes)){
		#col.def = paste(col.def,attributes[i],rpgsql.data.type(data.types[i]),options[i])
		col.def = paste(col.def,attributes[i],data.types[i],options[i])
		if(i!=length(attributes)){
			col.def <- paste(col.def,",")
		}
	}
	col.def <- paste(col.def,")")   
	query <- paste(query,col.def)
	# creating the table
	res <- dbSendQuery(conn,query)
	resInfo <- dbResultInfo(res)
	if(resInfo$error.message!=""){
		if(!no.transaction){
			dbSendQuery(conn,"ROLLBACK TRANSACTION;")
			if(do.log){
				loggit(res,call=paste("createDBTable: ERROR: creating table...",query))
			}
			stop("An error occurred! See the log file (by default tkfiDB.log in the working directory)!")
		}
	}
	else{
		if(!no.transaction){
			dbSendQuery(conn,"COMMIT TRANSACTION;")
		}
	}
	if(do.log){
		loggit(res,call=paste("createDBTable: INFO: creating table...",query))
	}
}


.insertIntoTable.old <- function(con,name=NULL,data=NULL,attributes=NULL,references=NULL,no.transaction=FALSE){
	if(is.null(name)){
		stop("Please submit the name of the table where you want to insert the data! \n\n")
	}
	if(is.null(data)){
		stop("No data to insert into the table! Remember that either the column names in the data table \n you submit, or by defining these names with the attributes \n you can specify the table columns (attributes) where the data should be inserted! \n\n")
	}
	if(!is.null(references)){
		if(class(references)=="list"){
			if(class(references[[1]])!="AutoReference"){
				stop("As reference parameter an object from the type AutoReference or a list of AutoReferences should be submitted! \n\n")
			}
		}
		else if(class(references)!="AutoReference"){
			stop("As reference parameter an object from the type AutoReference or a list of AutoReferences should be submitted! \n\n")
		}
	}
	# go on!
	if (!name %in% dbListTables(con)$relname)
		stop(paste("Table", name, "does not exist"))
	
	#dbSendQuery(con, "BEGIN TRANSACTION")
	#on.exit(if(!no.transaction){dbSendQuery(con, "ROLLBACK TRANSACTION")})
	#body
	if(is.null(attributes)){
		attribute.names <- colnames(data)
		#
		# do NOT remove or set this parameter. attributes should define the data columns in the original data.
		attributes <- colnames(data)
		if(is.null(attribute.names)){
			stop("Ups, perhaps you forgot to submit the column.names, or your data submitted has no column names specified.\n\n")
		}
	}
	else{
		attribute.names <- attributes
	}
	query <- paste("INSERT INTO",double.quote(name))
	attributes <- attribute.names
	# resolving the references
	ref.list <- ""
	ref.query <- ""
	if(!is.null(references)){
		if(class(references)!="list"){
			references <- list(references)
		}
		for(i in 1:length(references)){
				ref.list <- c(ref.list,getSourceAttribute(references[[i]]))
				attribute.names <- attribute.names[attribute.names!=getSourceColumn(references[[i]])]
				ref.query <- paste("(SELECT",getRefAttribute(references[[i]]),"FROM",getRefTable(references[[i]]),"WHERE",getRefColumn(references[[i]]))
		}
		attribute.names <- c(ref.list,attribute.names)
	}
	#######################################################
	query <- paste(query," (",list.to.csv(attribute.names[attribute.names!=""]),") VALUES ( ",sep="")
	
	
	data.values <- .toCharacterMatrix(data,one.col=(length(attributes)==1))
#	if(is.null(dim(data))){
#		nr.rows <- 1
#		data.values <- matrix(ncol=length(data),nrow=1)
#		data.values[1,] <- data
#	}
#	else{
#		data.values <- matrix(ncol=length(data[1,]),nrow=length(data[,1]))
#		for(i in 1:length(data[1,])){
#			data.values[,i] <- as.character(data[,i])
#		}
#		# can not use as.matrix, because for numbers it inserts blank spaces.
#		#data.values <- as.matrix(data)
#	}
	
	#colnames(data.values) <- attributes
	if(!no.transaction){
		dbSendQuery(con,"BEGIN TRANSACTION;")
	}
	error <- FALSE
	formatted.row <- NULL
#	data.values <- apply(data.values,2,rpgsql.format.values)
#	if(is.null(dim(data.values))){
#		nr.rows <- 1
#		data.v <- matrix(ncol=length(data.values),nrow=1)
#		data.v[1,] <- data.values
#		data.values <- data.v
#	}

	for(i in 1:nrow(data.values)){
		formatted.row <- rpgsql.format.values(data.values[i,])
		# resolve references, getting the primary key from the other tables and inserting them as foreign keys...
		if(!is.null(references)){
			new.row <- NULL
			bm <- rep(FALSE,length(attributes))
			if(class(references)!="list"){
				references <- list(references)
			}
			for(j in 1:length(references)){
					bm <- bm | attributes==getSourceColumn(references[[j]])
					formatted.row[attributes==getSourceColumn(references[[j]])] <- paste("(SELECT",getRefAttribute(references[[j]]),"FROM",getRefTable(references[[j]]),"WHERE",getRefColumn(references[[j]]),"=",formatted.row[attributes==getSourceColumn(references[[j]])],")")
					new.row <- c(new.row,formatted.row[attributes==getSourceColumn(references[[j]])])
			}
			formatted.row <- c(new.row,formatted.row[!bm])
		}
		res <- dbSendQuery(con,query,list.to.csv(formatted.row),");")
		if(do.log){
			if(loggit(res,call=paste("insertIntoTable: ",paste(query,list.to.csv(formatted.row),");")))){
				if(!no.transaction){
					dbSendQuery(con,"ROLLBACK TRANSACTION;")
				}
				error <- TRUE
				stop(paste("ERROR, an error occurred during the call: ",query))
			}
		}
		else{
			resultInfo <- dbResultInfo(res)
			if(resultInfo$error.message!=""){
				if(!no.transaction){
					dbSendQuery(con,"ROLLBACK TRANSACTION;")
				}
				error <- TRUE
				stop(paste("ERROR, an error occurred during the call: ",query))
			}
		}
	}
	if(!error){
		if(!no.transaction){
			dbSendQuery(con,"COMMIT TRANSACTION;")
		}
	}
	no.transaction=TRUE
	#on.exit(dbSendQuery(con, "COMMIT TRANSACTION"))
}



###########################################################################
insertIntoTable <- function(con,name=NULL,data=NULL,attributes=NULL,references=NULL,no.transaction=FALSE){
	if(is.null(name)){
		stop("Please submit the name of the table where you want to insert the data! \n\n")
	}
	if(is.null(data)){
		stop("No data to insert into the table! Remember that either the column names in the data table \n you submit, or by defining these names with the attributes \n you can specify the table columns (attributes) where the data should be inserted! \n\n")
	}
	if(!is.null(references)){
		if(class(references)=="list"){
			if(class(references[[1]])!="AutoReference"){
				stop("As reference parameter an object from the type AutoReference or a list of AutoReferences should be submitted! \n\n")
			}
		}
		else if(class(references)!="AutoReference"){
			stop("As reference parameter an object from the type AutoReference or a list of AutoReferences should be submitted! \n\n")
		}
	}
	# go on!
	if (!name %in% dbListTables(con)$relname)
		stop(paste("Table", name, "does not exist"))
	
	#dbSendQuery(con, "BEGIN TRANSACTION")
	#on.exit(if(!no.transaction){dbSendQuery(con, "ROLLBACK TRANSACTION")})
	#body
	if(is.null(attributes)){
		attribute.names <- colnames(data)
		#
		# do NOT remove or set this parameter. attributes should define the data columns in the original data.
		attributes <- colnames(data)
		if(is.null(attribute.names)){
			stop("Ups, perhaps you forgot to submit the column.names, or your data submitted has no column names specified.\n\n")
		}
	}
	else{
		attribute.names <- attributes
	}
	query <- paste("INSERT INTO",double.quote(name))
	attributes <- attribute.names
	# resolving the references
	ref.list <- ""
	ref.query <- ""
	if(!is.null(references)){
		if(class(references)!="list"){
			references <- list(references)
		}
		for(i in 1:length(references)){
				ref.list <- c(ref.list,getSourceAttribute(references[[i]]))
				attribute.names <- attribute.names[attribute.names!=getSourceColumn(references[[i]])]
				ref.query <- paste("(SELECT",getRefAttribute(references[[i]]),"FROM",getRefTable(references[[i]]),"WHERE",getRefColumn(references[[i]]))
		}
		attribute.names <- c(ref.list,attribute.names)
	}
	#######################################################
	query <- paste(query," (",list.to.csv(attribute.names[attribute.names!=""]),") VALUES ( ",sep="")
	
	
	data.values <- .toCharacterMatrix(data,one.col=(length(attributes)==1))
	
	if(!no.transaction){
		dbSendQuery(con,"BEGIN TRANSACTION;")
	}
	iserror <- FALSE
	formatted.row <- NULL
	iserror <- tryCatch(apply(data.values,MARGIN=1,.processInsert,con=con,references=references,no.transaction=no.transaction,query=query,attributes=attributes),interrupt=function(interrupt){
			.processError(con,no.transaction)
			return(TRUE)
			},
	warning=function(warning){},
	error=function(error){
			.processError(con,no.transaction)
			cat(as.character(error))
			return(TRUE)
			}
	)
	if(!is.null(class(iserror)) & !is.null(iserror)){
		if(class(iserror)=="list"){
			iserror = unlist(iserror)
		}
		iserror <- iserror[!is.na(iserror)]
		iserror = max(iserror)==1
		if(!iserror){
			if(!no.transaction){
				dbSendQuery(con,"COMMIT TRANSACTION;")
			}
		}
	}
	no.transaction=TRUE
}


.processError <- function(Con,no.transaction){
	#if(exists("ConnectionForErrors",envir=globalenv())){
	if(!no.transaction & dbConnectionInfo(Con)$status==0){
		dbSendQuery(Con,"ROLLBACK TRANSACTION")
		cat("\nThe transaction was rolled back\n")
	}	
	#}
}


.processInsert <- function(x,con,references=NULL,no.transaction=FALSE,query,attributes){
		formatted.row <- rpgsql.format.values(x)
		# resolve references, getting the primary key from the other tables and inserting them as foreign keys...
		if(!is.null(references)){
			new.row <- NULL
			bm <- rep(FALSE,length(attributes))
			if(class(references)!="list"){
				references <- list(references)
			}
			for(j in 1:length(references)){
					bm <- bm | attributes==getSourceColumn(references[[j]])
					formatted.row[attributes==getSourceColumn(references[[j]])] <- paste("(SELECT",getRefAttribute(references[[j]]),"FROM",getRefTable(references[[j]]),"WHERE",getRefColumn(references[[j]]),"=",formatted.row[attributes==getSourceColumn(references[[j]])],")")
					new.row <- c(new.row,formatted.row[attributes==getSourceColumn(references[[j]])])
			}
			formatted.row <- c(new.row,formatted.row[!bm])
		}
		res <- NULL
		res <- dbSendQuery(con,query,list.to.csv(formatted.row),");")
		if(!is.null(res)){
			if(loggit(res,call=paste("insertIntoTable: ",query,list.to.csv(formatted.row),");"))){
				stop(paste("ERROR, an error occurred during the call: ",query,list.to.csv(formatted.row),");"))
			}
		}
		return(FALSE)
}
###########################################################################






#############################################################################################################
#############################################################################################################
createSequence <- function(con,name=NULL,increment=1,start=100){
	if(!max(class(con)=="PgSQL.conn")==1){
		stop("The Connection you submitted was not initialized, it contains no PgSQL.conn object in the @con slot! \n Please use the dbConnect() method from the RdbiPgSQL package and create a PgSQLConnection object by submitting this object. \n\n")
	}
	if(is.null(name)){
		stop("Please submit the name of the sequence to be created! \n\n")
	}
	logMessage <- paste(" createSequencePG(con,name=",name,",",increment,",",start,")",sep="")
	query <- paste("CREATE SEQUENCE",double.quote(name),"INCREMENT BY",increment,"START",start,";")
	result <- dbSendQuery(con,query)
	if(do.log){
		loggit(result,call=logMessage)
	}
	if(dbResultInfo(result)$error.message!=""){
		stop(paste("An error occurred!! ",dbResultInfo(result)$status.string),dbResultInfo(result)$error.message,"\n\n")
	}
	result
}



############################################################################################################
############################################################################################################
loggit <- function(resultSet,call=""){
	ret <- FALSE
	resultInfo <- dbResultInfo(resultSet)
	if(log.level=="DEBUG" & do.log){
		#log anyway
		if(resultInfo$error.message!=""){
			call <- paste("ERROR: ",call)
			ret <- TRUE
		}
		message <- paste(">",date(),":",call,":",resultInfo$status.string,resultInfo$error.message,"rows=",resultInfo$rows,"cols=",resultInfo$columns,"response",resultInfo$command.response)
		write(file=log.file,message,append=TRUE)
	}
	else{
		if(resultInfo$error.message!=""){
		ret <- TRUE
		message <- paste(">",date(),": ERROR:  ",call,":",resultInfo$status.string,resultInfo$error.message,"rows=",resultInfo$rows,"cols=",resultInfo$columns,"response",resultInfo$command.response)
		write(file=log.file,message,append=TRUE)
		}
	}

	return(ret)
}







##########################################################################################################
##########################################################################################################
auto.pk.name <- function(name){
	name <- paste(name,"_pk",sep="")
	name
}

auto.pk.seq <- function(name){
	name <- paste(name,"_seq",sep="")
	name
}

auto.fk.name <- function(name){
	name <- paste(name,"_fk",sep="")
	name
}




################################################################################################################################
################################################################################################################################
setClass("AutoReference",representation(source.table="character",ref.table="character",source.table.column="character",ref.table.column="character"),
	prototype=list(source.table=NULL,ref.table=NULL,source.table.column=NULL,ref.table.column=NULL))

if( !isGeneric("getRefTable") )
	setGeneric("getRefTable", function(object)
	standardGeneric("getRefTable"))
		

setMethod("getRefTable","AutoReference",
	function(object){
		ret <- object@ref.table
		ret
	})
	
	
if( !isGeneric("getSourceColumn") )
	setGeneric("getSourceColumn", function(object)
	standardGeneric("getSourceColumn"))
		
setMethod("getSourceColumn","AutoReference",
	function(object){
		ret <- object@source.table.column
		ret
	})

if( !isGeneric("getRefColumn") )
	setGeneric("getRefColumn", function(object)
	standardGeneric("getRefColumn"))
		
setMethod("getRefColumn","AutoReference",
	function(object){
		ret <- object@ref.table.column
		ret
	})


if( !isGeneric("getRefAttribute") )
	setGeneric("getRefAttribute", function(object)
	standardGeneric("getRefAttribute"))
		

setMethod("getRefAttribute","AutoReference",
	function(object){
		ret <- auto.pk.name(object@ref.table)
		ret
	})


if( !isGeneric("getSourceAttribute") )
	setGeneric("getSourceAttribute", function(object)
	standardGeneric("getSourceAttribute"))
		

setMethod("getSourceAttribute","AutoReference",
	function(object){
		ret <- auto.fk.name(object@ref.table)
		ret
	})

	
if( !isGeneric("show") )
	setGeneric("show", function(object)
	standardGeneric("show"))
		

setMethod("show","AutoReference",
	function(object){
		cat("An object from the type AutoReference\n\n")
		cat(paste("Source table: \t\t",object@source.table,"\n"))
		cat(paste("Referenced table:\t",object@ref.table,"\n"))
		cat(paste("Attribute in the source table (that links to the referenced table):",getSourceAttribute(object),"\n"))
		cat(paste("Attribute in the referenced table where the attribute in the source table links to:",getRefAttribute(object),"\n\n"))
		cat(paste("Column in the data you submit thats data is the same as in the referenced table column (attribute):",getSourceColumn(object),"\n"))
		cat(paste("Column in the referenced table where the value is equal to the value of the source data column (specified above):",getRefColumn(object),"\n\n"))
	})
	

	
	
############################################################################################################################
############################################################################################################################
dbListSequences <- function(conn, pattern = NULL, all = FALSE) {

  query <- paste("SELECT relname FROM pg_class WHERE relkind = 'S'")

  if (!is.null(pattern))
    query <- paste(query, "AND relname ~", single.quote(pattern))
  # match by regular expression

  if (!all)
    query <- paste(query, "AND NOT relname ~ '^pg_'")
  # match by regular expression

  query <- paste(query, "ORDER BY relname")

  tables <- dbGetQuery(conn, query)

  return(tables)

}



#####################################################################################################################################
#####################################################################################################################################
.updateDBTable.old <- function(con=NULL,name=NULL,set.col=NULL,where.col=NULL,new.values=NULL,old.values=NULL,condition="=",references=NULL,no.transaction=FALSE){
	if(is.null(con)){
		stop("no connection submitted! \n\n")
	}
	if(is.null(name)){
		stop("please submit the table name. \n\n")
	}
	if(is.null(set.col)){
		if(is.null(colnames(new.values))){
			stop("please define the column where the values shoudl be updated with the set.col parameter (or the colnames of the new.values if a matrix or data frame was submitted). \n\n")
		}
		else{
			set.col = colnames(new.values)
		}
	}
	if(is.null(where.col)){
		if(is.null(colnames(old.values))){
			stop("please define the column where the condition should be applied with the where.col parameter.\n Alternatively you can specify these as column names in the old.values parmeter \n\n")
		}
		else{
			where.col = colnames(old.values)
		}
	}
	if(is.null(new.values)){
		stop("please submit the new values (new.values). \n\n")
	}
	if(is.null(old.values)){
		stop("please submit the old values (old.values). The function queries the where.col with these values to define the rows to update. \n\n")
	}
	if(!is.null(references)){
		if(class(references)=="list"){
			if(class(references[[1]])!="AutoReference"){
				stop("As reference parameter an object from the type AutoReference or a list of AutoReferences should be submitted! \n\n")
			}
		}
		else if(class(references)!="AutoReference"){
			stop("As reference parameter an object from the type AutoReference or a list of AutoReferences should be submitted! \n\n")
		}
	}
	# error handling finished.
	
#	if(length(new.values)!=length(old.values)){
#		stop("Length of old.values and new.values differ... \n\n")
#		
#	}
	
	error <- FALSE
	# to be flexible makes everithing more complicated...
	if(length(set.col)>1){
		if(is.null(dim(new.values))){
			mat <- matrix(ncol=length(set.col),nrow=1)
			mat[1,] <- new.values
			new.values <- mat
		}
		else{
			if(length(new.values[1,])!=length(set.col)){
				stop("length of the parameter set.col and the number of columns in the data new.values sumbitted differ!\n\n")
			}
		}
	}
	if(length(where.col)>1){
		if(is.null(dim(old.values))){
			mat <- matrix(ncol=length(where.col),nrow=1)
			mat[1,] <- old.values
			old.values <- mat
		}
		else{
			if(length(old.values[1,])!=length(where.col)){
				stop("length of the parameter set.col and the number of columns in the data old.values sumbitted differ!\n\n")
			}
		}
	}
	#####
	
	
	
#	old.values <- as.matrix(old.values)
#	new.values <- as.matrix(new.values)
	old.values <- .toCharacterMatrix(old.values,one.col=(length(where.col)==1))
	new.values <- .toCharacterMatrix(new.values,one.col=(length(set.col)==1))
	
	for(i in 1:length(old.values[,1])){
		old.values[i,] <- rpgsql.format.values(format(old.values[i,]))
		new.values[i,] <- rpgsql.format.values(format(new.values[i,]))
	}
		
	# the number of rows must be the same!!!
	if(length(old.values[,1])!=length(new.values[,1])){
		stop("I'm sorry, but the number of rows in new.values must be the same as the number of rows in old.values!!\n\n")
	}
	# removing columns from old.values that are needed by the references
	if(!is.null(references)){
		if(class(references)!="list"){
			references <- list(references)
		}
		ref.table.values <- matrix(ncol=length(references),nrow=length(old.values[,1]))
		new.values.n <- matrix(ncol=(length(new.values[1,])+length(references)),nrow=length(new.values[,1]))
		new.values.n[,1:(length(new.values[1,]))] <- new.values
		new.values <- new.values.n
		col.names <- NULL
		for(i in 1:length(references)){
			# dreating the column names for the new ref table
			col.names <- c(col.names,getSourceColumn(references[[i]]))
			# filling the ref table with values
			ref.table.values[,i] <- old.values[,where.col==getSourceColumn(references[[i]])]
			# naturally have to remove these columns out of the original table
			old.values <- old.values[,where.col!=getSourceColumn(references[[i]])]
			if(is.null(dim(old.values))){
				old <- matrix(ncol=1,nrow=length(old.values))
				old[,1] <- old.values
				old.values <- old
			}
			where.col <- where.col[where.col!=getSourceColumn(references[[i]])]
			
			# adding the column to set
			set.col <- c(set.col,getSourceAttribute(references[[i]]))
			# generating the SELECT queries for the foreign keys
			for(j in 1:length(new.values[,1])){
				new.values[j,(length(set.col))] <- paste("(SELECT",getRefAttribute(references[[i]]),"FROM",getRefTable(references[[i]]),"WHERE",getRefColumn(references[[i]]),"=",(ref.table.values[j,i]),")")
			}
		}
	}
	if(!no.transaction){
		dbSendQuery(con,"BEGIN TRANSACTION;")
	}
	for(i in 1:length(old.values[,1])){
		#current.old <- rpgsql.format.values(old.values[i,])
		#current.new <- rpgsql.format.values(new.values[i,])

		current.old <- (old.values[i,])
		current.new <- (new.values[i,])

		query <- .generateUpdateQuery(name,set.col,current.new,where.col,current.old,condition=condition)
		res <- dbSendQuery(con,query)
		if(do.log){
			loggit(res,call=paste("updateDBTable:",query))
		}
		resInfo <- dbResultInfo(res)
		if(resInfo$error.message!=""){
			if(!no.transaction){
				dbSendQuery(con,"ROLLBACK TRANSACTION;")
			}
			error <- TRUE
			stop(paste("ERROR, an error occurred during the call: ",query))
		}
	}
	if(!error){
		if(!no.transaction){
			dbSendQuery(con,"COMMIT TRANSACTION;")
		}
	}
	no.transaction=TRUE
	#if(!no.transaction){
	#	on.exit/java(dbSendQuery(con, "ROLLBACK TRANSACTION"))
	#}
}


#########################################################################################################################
updateDBTable <- function(con=NULL,name=NULL,set.col=NULL,where.col=NULL,new.values=NULL,old.values=NULL,condition="=",references=NULL,no.transaction=FALSE){
	if(is.null(con)){
		stop("no connection submitted! \n\n")
	}
	if(is.null(name)){
		stop("please submit the table name. \n\n")
	}
	if(is.null(set.col)){
		if(is.null(colnames(new.values))){
			stop("please define the column where the values shoudl be updated with the set.col parameter (or the colnames of the new.values if a matrix or data frame was submitted). \n\n")
		}
		else{
			set.col = colnames(new.values)
		}
	}
	if(is.null(where.col)){
		if(is.null(colnames(old.values))){
			stop("please define the column where the condition should be applied with the where.col parameter.\n Alternatively you can specify these as column names in the old.values parmeter \n\n")
		}
		else{
			where.col = colnames(old.values)
		}
	}
	if(is.null(new.values)){
		stop("please submit the new values (new.values). \n\n")
	}
	if(is.null(old.values)){
		stop("please submit the old values (old.values). The function queries the where.col with these values to define the rows to update. \n\n")
	}
	if(!is.null(references)){
		if(class(references)=="list"){
			if(class(references[[1]])!="AutoReference"){
				stop("As reference parameter an object from the type AutoReference or a list of AutoReferences should be submitted! \n\n")
			}
		}
		else if(class(references)!="AutoReference"){
			stop("As reference parameter an object from the type AutoReference or a list of AutoReferences should be submitted! \n\n")
		}
	}
	# error handling finished.
	
#	if(length(new.values)!=length(old.values)){
#		stop("Length of old.values and new.values differ... \n\n")
#		
#	}
	
	error <- FALSE
	# to be flexible makes everithing more complicated...
	if(length(set.col)>1){
		if(is.null(dim(new.values))){
			mat <- matrix(ncol=length(set.col),nrow=1)
			mat[1,] <- new.values
			new.values <- mat
		}
		else{
			if(length(new.values[1,])!=length(set.col)){
				stop("length of the parameter set.col and the number of columns in the data new.values sumbitted differ!\n\n")
			}
		}
	}
	if(length(where.col)>1){
		if(is.null(dim(old.values))){
			mat <- matrix(ncol=length(where.col),nrow=1)
			mat[1,] <- old.values
			old.values <- mat
		}
		else{
			if(length(old.values[1,])!=length(where.col)){
				stop("length of the parameter set.col and the number of columns in the data old.values sumbitted differ!\n\n")
			}
		}
	}
	#####
	
	
	
#	old.values <- as.matrix(old.values)
#	new.values <- as.matrix(new.values)
	old.values <- .toCharacterMatrix(old.values,one.col=(length(where.col)==1))
	new.values <- .toCharacterMatrix(new.values,one.col=(length(set.col)==1))
	
#	for(i in 1:length(old.values[,1])){
#		old.values[i,] <- rpgsql.format.values(format(old.values[i,]))
#		new.values[i,] <- rpgsql.format.values(format(new.values[i,]))
#	}
		
	# the number of rows must be the same!!!
	if(length(old.values[,1])!=length(new.values[,1])){
		stop("I'm sorry, but the number of rows in new.values must be the same as the number of rows in old.values!!\n\n")
	}
	# removing columns from old.values that are needed by the references
	if(!is.null(references)){
		if(class(references)!="list"){
			references <- list(references)
		}
		ref.table.values <- matrix(ncol=length(references),nrow=length(old.values[,1]))
		new.values.n <- matrix(ncol=(length(new.values[1,])+length(references)),nrow=length(new.values[,1]))
		new.values.n[,1:(length(new.values[1,]))] <- new.values
		new.values <- new.values.n
		col.names <- NULL
		for(i in 1:length(references)){
			# dreating the column names for the new ref table
			col.names <- c(col.names,getSourceColumn(references[[i]]))
			# filling the ref table with values
			ref.table.values[,i] <- old.values[,where.col==getSourceColumn(references[[i]])]
			# naturally have to remove these columns out of the original table
			old.values <- old.values[,where.col!=getSourceColumn(references[[i]])]
			if(is.null(dim(old.values))){
				old <- matrix(ncol=1,nrow=length(old.values))
				old[,1] <- old.values
				old.values <- old
			}
			where.col <- where.col[where.col!=getSourceColumn(references[[i]])]
			
			# adding the column to set
			set.col <- c(set.col,getSourceAttribute(references[[i]]))
			# generating the SELECT queries for the foreign keys
			for(j in 1:length(new.values[,1])){
				new.values[j,(length(set.col))] <- paste("(SELECT",getRefAttribute(references[[i]]),"FROM",getRefTable(references[[i]]),"WHERE",getRefColumn(references[[i]]),"=",(ref.table.values[j,i]),")")
			}
		}
	}
	if(!no.transaction){
		dbSendQuery(con,"BEGIN TRANSACTION;")
	}
	big.table <- cbind(old.values,new.values)
	old.columns <- rep(FALSE,ncol(big.table))
	old.columns[1:ncol(old.values)] <- TRUE
	error <- tryCatch(apply(big.table,MARGIN=1,FUN=.processUpdate,con=con,old.columns=old.columns,name=name,set.col=set.col,where.col=where.col,condition=condition),interrupt=function(interrupt){
		.processError(con,no.transaction)
		return(TRUE)
		},
	error=function(error){
		.processError(con,no.transaction)
		return(TRUE)
	})
	if(!is.null(class(error)) & !is.null(error)){
		if(class(error)=="list"){
			error = unlist(error)
		}
		error <- error[!is.na(error)]
		error <- max(error)==1
		if(!error){
			if(!no.transaction){
				dbSendQuery(con,"COMMIT TRANSACTION;")
			}
		}
	}
	no.transaction=TRUE
}
#########################################################################################################################

.processUpdate <- function(x,con,old.columns,name,set.col,where.col,condition){
	current.old <- rpgsql.format.values(x[old.columns])
	current.new <- rpgsql.format.values(x[!old.columns])
	query <- .generateUpdateQuery(name,set.col,current.new,where.col,current.old,condition=condition)
	res <- NULL
	res <- dbSendQuery(con,query)
	if(!is.null(res)){
		if(loggit(res,call=paste("updateDBTable: ",query,");"))){
				stop(paste("ERROR, an error occurred during the call: ",query))
		}
	}
	return(FALSE)
}

#########################################################################################################################
#########################################################################################################################
.generateUpdateQuery <- function(table,set,new.value,where,old.value,condition="="){
	if(length(condition)!=length(where)){
		condition <- rep(condition[1],length(where))
	}
	set.query = NULL
	where.query = ""
	for(i in 1:length(set)){
		set.query = c(set.query,paste(double.quote(set[i]),"=",new.value[i]))
	}
	for(i in 1:length(where)){
		if(i==length(where)){
			and = ""
		}
		else{
			and = "AND"
		}
		where.query = paste(where.query,double.quote(where[i]),condition[i],old.value[i],and)
	}
	query <- paste("UPDATE ",double.quote(table),"SET",list.to.csv(set.query),"WHERE",where.query,";")
	query
}


#########################################################################################################################
## deleting rows from a database table
#########################################################################################################################
deleteDBEntry <- function(con,name=NULL,attribute=NULL,value=NULL,condition="=",no.transaction=FALSE){
	if( !(name %in% dbListTables(con)$relname)){
		stop(paste("Table",name,"is not a table in your database!\n\n"))
	}
	if(!is.null(value)){
		if(is.null(attribute)){
			if(is.null(colnames(value))){
				stop("Plese submit the column where the value that you submitted should be compared to with the attribute argument! \n\n")
			}
			else{
				attribute <- colnames(value)[1]
			}
		}
	}
	
	if(!no.transaction){
		dbSendQuery(con,"BEGIN TRANSACTION;")
	}
	query <- paste("DELETE FROM",double.quote(name))
	
	if(!is.null(value)){
		for(i in 1:length(value)){
			where <- paste("WHERE",attribute,condition,rpgsql.format.values(value[i]),";")
			res <- dbSendQuery(con,paste(query,where,";"))
			if(do.log){
				loggit(res,call=paste("deleteDBEntry;",query,where))
			}
		}
	}
	else{
		cat(paste("Deleting all entries in table",name))
		res <- dbSendQuery(con,query,";")
		if(do.log){
			loggit(res,call=paste("deleteDBEntry:",query))
		}
	}
		
	if(!no.transaction){
		dbSendQuery(con,"COMMIT TRANSACTION")
	}
	no.transaction=TRUE
}

#pgsql.format.values(){
#}

##
# function that converts a matrix to a character matrix 
.toCharacterMatrix <- function(data,one.col=TRUE){
	if(is.null(dim(data))){
		#nr.rows <- 1
		if(one.col){
			data.values <- matrix(nrow=length(data),ncol=1)
			data.values[,1] <- data
		}
		else{
			data.values <- matrix(ncol=length(data),nrow=1)
			data.values[1,] <- data
		}
	}
	else{
		data.values <- apply(data,c(1,2),as.character)
		#data.values <- matrix(ncol=length(data[1,]),nrow=length(data[,1]))
		#for(i in 1:length(data[1,])){
		#	data.values[,i] <- as.character(data[,i])
		#}
	}
		data.values
}


##
# Search in a database table...
searchIn <- function(Con,name=NULL,pattern=NULL,where=NULL){
	if(is.null(name)){
		stop("Please submit the table name where the string should be searched! \n\n")
	}
	if(is.null(pattern)){
		stop("hm, strange, you do not want to search something? You forgot to submit the text that should be searched...\n\n")
	}
	if(is.null(where)){
		res <- dbSendQuery(Con,"SELECT * FROM ",name,"LIMIT 1;")
		resInfo <- dbResultInfo(res)
		if(resInfo$error.message!=""){
			if(do.log){
				loggit(res,call=paste("selectIn: ERROR: ","SELECT * FROM ",name,"LIMIT 1;"))
			}
				stop("An error occurred! See the log file (by default tkfiDB.log in the working directory)!")
		}
		where <- colnames(dbGetResult(res))
		##
		# reduce the search to those columns that have the SQL type TEXT.
		where <- where[dbColumnInfo(res)[,"Type"]==25]
	}
	where.query <- ""
	for(i in 1:length(where)){
		where.query <- paste(where.query,where[i],"LIKE",rpgsql.format.values(paste("%",pattern,"%",sep="")))
		if(i < length(where)){
			where.query <- paste(where.query,"OR")
		}
	}
	res <- dbSendQuery(Con,"SELECT * FROM",name,"WHERE",where.query,";")
	resInfo <- dbResultInfo(res)
	if(resInfo$error.message!=""){
		if(do.log){
			loggit(res,call=paste("selectIn: ERROR: ","SELECT * FROM",name,"WHERE",where.query,";"))
		}
			stop("An error occurred! See the log file (by default tkfiDB.log in the working directory)!")
	}
	if(do.log){
		loggit(res,call=paste("createPCGTable: INFO: ","SELECT * FROM",name,"WHERE",where.query,";"))
	}
	result <- dbGetResult(res)
	result
}


dbColnames <- function(con,table.name){
	result <- dbSendQuery(con,"SELECT * FROM",table.name,"LIMIT 1")
	resInfo <- dbResultInfo(result)
	if(resInfo$error.message!=""){
		if(do.log){
			loggit(res,call=paste("dbColnames: ERROR: ","SELECT * FROM",table.name,"LIMIT 1"))
		}
		stop("An error occurred! See the log file (by default tkfiDB.log in the working directory)!")
	}
	Colnames <- colnames(dbGetResult(result))
	Colnames
}


dbListDatabases <- function(con){
	res <- dbSendQuery(con,"SELECT * FROM pg_database")
	result <- dbGetResult(res)
	result
}



